@@ -34,7 +34,7 @@
  *    are probably inadequate.
  *
  * Loop detection based on AS_PATH causes updates to be withdrawn. RFC
- * 4271 does not explicitly specify the behavior in that case.
+ * 4271 does not explicitly specifiy the behavior in that case.
  *
  * Loop detection related to route reflection (based on ORIGINATOR_ID
  * and CLUSTER_LIST) causes updates to be withdrawn. RFC 4456 8
@@ -45,9 +45,9 @@
  *
  * export - Hook that validates and normalizes attribute during export phase.
  * Receives eattr, may modify it (e.g., sort community lists for canonical
- * representation), UNSET() it (e.g., skip empty lists), or REJECT() the route
- * if necessary. May assume that eattr has value valid w.r.t. its type, but may
- * be invalid w.r.t. BGP constraints. Optional.
+ * representation), UNSET() it (e.g., skip empty lists), or WITHDRAW() it if
+ * necessary. May assume that eattr has value valid w.r.t. its type, but may be
+ * invalid w.r.t. BGP constraints. Optional.
  *
  * encode - Hook that converts internal representation to external one during
  * packet writing. Receives eattr and puts it in the buffer (including attribute
@@ -72,7 +72,7 @@ struct bgp_attr_desc {
   void (*export)(struct bgp_export_state *s, eattr *a);
   int  (*encode)(struct bgp_write_state *s, eattr *a, byte *buf, uint size);
   void (*decode)(struct bgp_parse_state *s, uint code, uint flags, byte *data, uint len, ea_list **to);
-  void (*format)(const eattr *ea, byte *buf, uint size);
+  void (*format)(eattr *ea, byte *buf, uint size);
 };
 
 static const struct bgp_attr_desc bgp_attr_table[];
@@ -88,7 +88,7 @@ bgp_set_attr(ea_list **attrs, struct linpool *pool, uint code, uint flags, uintp
       attrs,
       pool,
       EA_CODE(PROTOCOL_BGP, code),
-      flags & ~BAF_EXT_LEN,
+      flags,
       bgp_attr_table[code].type,
       val
   );
@@ -108,9 +108,6 @@ bgp_set_attr(ea_list **attrs, struct linpool *pool, uint code, uint flags, uintp
 #define UNSET(a) \
   ({ a->type = EAF_TYPE_UNDEF; return; })
 
-#define REJECT(msg, args...) \
-  ({ log(L_ERR "%s: " msg, s->proto->p.name, ## args); s->err_reject = 1; return; })
-
 #define NEW_BGP		"Discarding %s attribute received from AS4-aware neighbor"
 #define BAD_EBGP	"Discarding %s attribute received from EBGP neighbor"
 #define BAD_LENGTH	"Malformed %s attribute - invalid length (%u)"
@@ -121,7 +118,7 @@ bgp_set_attr(ea_list **attrs, struct linpool *pool, uint code, uint flags, uintp
 static inline int
 bgp_put_attr_hdr3(byte *buf, uint code, uint flags, uint len)
 {
-  *buf++ = flags & ~BAF_EXT_LEN;
+  *buf++ = flags;
   *buf++ = code;
   *buf++ = len;
   return 3;
@@ -202,179 +199,6 @@ bgp_encode_raw(struct bgp_write_state *s UNUSED, eattr *a, byte *buf, uint size)
 }
 
 
-/*
- *	AIGP handling
- */
-
-static int
-bgp_aigp_valid(byte *data, uint len, char *err, uint elen)
-{
-  byte *pos = data;
-  char *err_dsc = NULL;
-  uint err_val = 0;
-
-#define BAD(DSC,VAL) ({ err_dsc = DSC; err_val = VAL; goto bad; })
-  while (len)
-  {
-    if (len < 3)
-      BAD("TLV framing error", len);
-
-    /* Process one TLV */
-    uint ptype = pos[0];
-    uint plen = get_u16(pos + 1);
-
-    if (len < plen)
-      BAD("TLV framing error", plen);
-
-    if (plen < 3)
-      BAD("Bad TLV length", plen);
-
-    if ((ptype == BGP_AIGP_METRIC) && (plen != 11))
-      BAD("Bad AIGP TLV length", plen);
-
-    ADVANCE(pos, len, plen);
-  }
-#undef BAD
-
-  return 1;
-
-bad:
-  if (err)
-    if (bsnprintf(err, elen, "%s (%u) at %d", err_dsc, err_val, (int) (pos - data)) < 0)
-      err[0] = 0;
-
-  return 0;
-}
-
-static const byte *
-bgp_aigp_get_tlv(const struct adata *ad, uint type)
-{
-  if (!ad)
-    return NULL;
-
-  uint len = ad->length;
-  const byte *pos = ad->data;
-
-  while (len)
-  {
-    uint ptype = pos[0];
-    uint plen = get_u16(pos + 1);
-
-    if (ptype == type)
-      return pos;
-
-    ADVANCE(pos, len, plen);
-  }
-
-  return NULL;
-}
-
-static const struct adata *
-bgp_aigp_set_tlv(struct linpool *pool, const struct adata *ad, uint type, byte *data, uint dlen)
-{
-  uint len = ad ? ad->length : 0;
-  const byte *pos = ad ? ad->data : NULL;
-  struct adata *res = lp_alloc_adata(pool, len + 3 + dlen);
-  byte *dst = res->data;
-  byte *tlv = NULL;
-  int del = 0;
-
-  while (len)
-  {
-    uint ptype = pos[0];
-    uint plen = get_u16(pos + 1);
-
-    /* Find position for new TLV */
-    if ((ptype >= type) && !tlv)
-    {
-      tlv = dst;
-      dst += 3 + dlen;
-    }
-
-    /* Skip first matching TLV, copy others */
-    if ((ptype == type) && !del)
-      del = 1;
-    else
-    {
-      memcpy(dst, pos, plen);
-      dst += plen;
-    }
-
-    ADVANCE(pos, len, plen);
-  }
-
-  if (!tlv)
-  {
-    tlv = dst;
-    dst += 3 + dlen;
-  }
-
-  /* Store the TLD */
-  put_u8(tlv + 0, type);
-  put_u16(tlv + 1, 3 + dlen);
-  memcpy(tlv + 3, data, dlen);
-
-  /* Update length */
-  res->length = dst - res->data;
-
-  return res;
-}
-
-static u64 UNUSED
-bgp_aigp_get_metric(const struct adata *ad, u64 def)
-{
-  const byte *b = bgp_aigp_get_tlv(ad, BGP_AIGP_METRIC);
-  return b ? get_u64(b + 3) : def;
-}
-
-static const struct adata *
-bgp_aigp_set_metric(struct linpool *pool, const struct adata *ad, u64 metric)
-{
-  byte data[8];
-  put_u64(data, metric);
-  return bgp_aigp_set_tlv(pool, ad, BGP_AIGP_METRIC, data, 8);
-}
-
-int
-bgp_total_aigp_metric_(rte *e, u64 *metric, const struct adata **ad)
-{
-  eattr *a = ea_find(e->attrs->eattrs, EA_CODE(PROTOCOL_BGP, BA_AIGP));
-  if (!a)
-    return 0;
-
-  const byte *b = bgp_aigp_get_tlv(a->u.ptr, BGP_AIGP_METRIC);
-  if (!b)
-    return 0;
-
-  u64 aigp = get_u64(b + 3);
-  u64 step = e->attrs->igp_metric;
-
-  if (!rte_resolvable(e) || (step >= IGP_METRIC_UNKNOWN))
-    step = BGP_AIGP_MAX;
-
-  if (!step)
-    step = 1;
-
-  *ad = a->u.ptr;
-  *metric = aigp + step;
-  if (*metric < aigp)
-    *metric = BGP_AIGP_MAX;
-
-  return 1;
-}
-
-static inline int
-bgp_init_aigp_metric(rte *e, u64 *metric, const struct adata **ad)
-{
-  if (e->attrs->source == RTS_BGP)
-    return 0;
-
-  *metric = rt_get_igp_metric(e);
-  *ad = NULL;
-  return *metric < IGP_METRIC_UNKNOWN;
-}
-
-
 /*
  *	Attribute hooks
  */
@@ -383,7 +207,7 @@ static void
 bgp_export_origin(struct bgp_export_state *s, eattr *a)
 {
   if (a->u.data > 2)
-    REJECT(BAD_VALUE, "ORIGIN", a->u.data);
+    WITHDRAW(BAD_VALUE, "ORIGIN", a->u.data);
 }
 
 static void
@@ -399,7 +223,7 @@ bgp_decode_origin(struct bgp_parse_state *s, uint code UNUSED, uint flags, byte
 }
 
 static void
-bgp_format_origin(const eattr *a, byte *buf, uint size UNUSED)
+bgp_format_origin(eattr *a, byte *buf, uint size UNUSED)
 {
   static const char *bgp_origin_names[] = { "IGP", "EGP", "Incomplete" };
 
@@ -407,15 +231,6 @@ bgp_format_origin(const eattr *a, byte *buf, uint size UNUSED)
 }
 
 
-static inline int
-bgp_as_path_first_as_equal(const byte *data, uint len, u32 asn)
-{
-  return (len >= 6) &&
-    ((data[0] == AS_PATH_SEQUENCE) || (data[0] == AS_PATH_CONFED_SEQUENCE)) &&
-    (data[1] > 0) &&
-    (get_u32(data+2) == asn);
-}
-
 static int
 bgp_encode_as_path(struct bgp_write_state *s, eattr *a, byte *buf, uint size)
 {
@@ -438,13 +253,17 @@ bgp_decode_as_path(struct bgp_parse_state *s, uint code UNUSED, uint flags, byte
 {
   struct bgp_proto *p = s->proto;
   int as_length = s->as4_session ? 4 : 2;
-  int as_sets = p->cf->allow_as_sets;
   int as_confed = p->cf->confederation && p->is_interior;
   char err[128];
 
-  if (!as_path_valid(data, len, as_length, as_sets, as_confed, err, sizeof(err)))
+  if (!as_path_valid(data, len, as_length, as_confed, err, sizeof(err)))
     WITHDRAW("Malformed AS_PATH attribute - %s", err);
 
+  /* In some circumstances check for initial AS_CONFED_SEQUENCE; RFC 5065 5.0 */
+  if (p->is_interior && !p->is_internal &&
+      ((len < 2) || (data[0] != AS_PATH_CONFED_SEQUENCE)))
+    WITHDRAW("Malformed AS_PATH attribute - %s", "missing initial AS_CONFED_SEQUENCE");
+
   if (!s->as4_session)
   {
     /* Prepare 32-bit AS_PATH (from 16-bit one) in a temporary buffer */
@@ -453,16 +272,6 @@ bgp_decode_as_path(struct bgp_parse_state *s, uint code UNUSED, uint flags, byte
     len = as_path_16to32(data, src, len);
   }
 
-  /* In some circumstances check for initial AS_CONFED_SEQUENCE; RFC 5065 5.0 */
-  if (p->is_interior && !p->is_internal &&
-      ((len < 2) || (data[0] != AS_PATH_CONFED_SEQUENCE)))
-    WITHDRAW("Malformed AS_PATH attribute - %s", "missing initial AS_CONFED_SEQUENCE");
-
-  /* Reject routes with first AS in AS_PATH not matching neighbor AS; RFC 4271 6.3 */
-  if (!p->is_internal && p->cf->enforce_first_as &&
-      !bgp_as_path_first_as_equal(data, len, p->remote_as))
-    WITHDRAW("Malformed AS_PATH attribute - %s", "First AS differs from neigbor AS");
-
   bgp_set_attr_data(to, s->pool, BA_AS_PATH, flags, data, len);
 }
 
@@ -513,7 +322,7 @@ bgp_decode_next_hop(struct bgp_parse_state *s, uint code UNUSED, uint flags UNUS
 
 /* TODO: This function should use AF-specific hook */
 static void
-bgp_format_next_hop(const eattr *a, byte *buf, uint size UNUSED)
+bgp_format_next_hop(eattr *a, byte *buf, uint size UNUSED)
 {
   ip_addr *nh = (void *) a->u.ptr->data;
   uint len = a->u.ptr->length;
@@ -580,7 +389,6 @@ bgp_encode_aggregator(struct bgp_write_state *s, eattr *a, byte *buf, uint size)
     /* Prepare 16-bit AGGREGATOR (from 32-bit one) in a temporary buffer */
     byte *dst = alloca(6);
     len = aggregator_32to16(dst, data);
-    data = dst;
   }
 
   return bgp_put_attr(buf, size, BA_AGGREGATOR, a->flags, data, len);
@@ -604,7 +412,7 @@ bgp_decode_aggregator(struct bgp_parse_state *s, uint code UNUSED, uint flags, b
 }
 
 static void
-bgp_format_aggregator(const eattr *a, byte *buf, uint size UNUSED)
+bgp_format_aggregator(eattr *a, byte *buf, uint size UNUSED)
 {
   const byte *data = a->u.ptr->data;
 
@@ -679,44 +487,13 @@ bgp_decode_cluster_list(struct bgp_parse_state *s, uint code UNUSED, uint flags,
 }
 
 static void
-bgp_format_cluster_list(const eattr *a, byte *buf, uint size)
+bgp_format_cluster_list(eattr *a, byte *buf, uint size)
 {
   /* Truncates cluster lists larger than buflen, probably not a problem */
   int_set_format(a->u.ptr, 0, -1, buf, size);
 }
 
 
-int
-bgp_encode_mp_reach_mrt(struct bgp_write_state *s UNUSED, eattr *a, byte *buf, uint size)
-{
-  /*
-   *	Limited version of MP_REACH_NLRI used for MRT table dumps (IPv6 only):
-   *
-   *	3 B	MP_REACH_NLRI header
-   *	1 B	MP_REACH_NLRI data - Length of Next Hop Network Address
-   *	var	MP_REACH_NLRI data - Network Address of Next Hop
-   */
-
-  ip_addr *nh = (void *) a->u.ptr->data;
-  uint len = a->u.ptr->length;
-
-  ASSERT((len == 16) || (len == 32));
-
-  if (size < (3+1+len))
-    return -1;
-
-  bgp_put_attr_hdr3(buf, BA_MP_REACH_NLRI, BAF_OPTIONAL, 1+len);
-  buf[3] = len;
-  buf += 4;
-
-  put_ip6(buf, ipa_to_ip6(nh[0]));
-
-  if (len == 32)
-    put_ip6(buf+16, ipa_to_ip6(nh[1]));
-
-  return 3+1+len;
-}
-
 static inline u32
 get_af3(byte *buf)
 {
@@ -767,23 +544,13 @@ bgp_decode_mp_unreach_nlri(struct bgp_parse_state *s, uint code UNUSED, uint fla
 static void
 bgp_export_ext_community(struct bgp_export_state *s, eattr *a)
 {
-  if (!s->proto->is_interior)
-  {
-    struct adata *ad = ec_set_del_nontrans(s->pool, a->u.ptr);
+  struct adata *ad = ec_set_del_nontrans(s->pool, a->u.ptr);
 
-    if (ad->length == 0)
-      UNSET(a);
+  if (ad->length == 0)
+    UNSET(a);
 
-    ec_set_sort_x(ad);
-    a->u.ptr = ad;
-  }
-  else
-  {
-    if (a->u.ptr->length == 0)
-      UNSET(a);
-
-    a->u.ptr = ec_set_sort(s->pool, a->u.ptr);
-  }
+  ec_set_sort_x(ad);
+  a->u.ptr = ad;
 }
 
 static void
@@ -813,9 +580,6 @@ bgp_decode_as4_aggregator(struct bgp_parse_state *s, uint code UNUSED, uint flag
 static void
 bgp_decode_as4_path(struct bgp_parse_state *s, uint code UNUSED, uint flags, byte *data, uint len, ea_list **to)
 {
-  struct bgp_proto *p = s->proto;
-  int sets = p->cf->allow_as_sets;
-
   char err[128];
 
   if (s->as4_session)
@@ -824,7 +588,7 @@ bgp_decode_as4_path(struct bgp_parse_state *s, uint code UNUSED, uint flags, byt
   if (len < 6)
     DISCARD(BAD_LENGTH, "AS4_PATH", len);
 
-  if (!as_path_valid(data, len, 4, sets, 1, err, sizeof(err)))
+  if (!as_path_valid(data, len, 4, 1, err, sizeof(err)))
     DISCARD("Malformed AS4_PATH attribute - %s", err);
 
   struct adata *a = lp_alloc_adata(s->pool, len);
@@ -840,42 +604,6 @@ bgp_decode_as4_path(struct bgp_parse_state *s, uint code UNUSED, uint flags, byt
   bgp_set_attr_ptr(to, s->pool, BA_AS4_PATH, flags, a);
 }
 
-
-static void
-bgp_export_aigp(struct bgp_export_state *s, eattr *a)
-{
-  if (!s->channel->cf->aigp)
-    UNSET(a);
-}
-
-static void
-bgp_decode_aigp(struct bgp_parse_state *s, uint code UNUSED, uint flags, byte *data, uint len, ea_list **to)
-{
-  char err[128];
-
-  /* Acceptability test postponed to bgp_finish_attrs() */
-
-  if ((flags ^ bgp_attr_table[BA_AIGP].flags) & (BAF_OPTIONAL | BAF_TRANSITIVE))
-    DISCARD("Malformed AIGP attribute - conflicting flags (%02x)", flags);
-
-  if (!bgp_aigp_valid(data, len, err, sizeof(err)))
-    DISCARD("Malformed AIGP attribute - %s", err);
-
-  bgp_set_attr_data(to, s->pool, BA_AIGP, flags, data, len);
-}
-
-static void
-bgp_format_aigp(const eattr *a, byte *buf, uint size UNUSED)
-{
-  const byte *b = bgp_aigp_get_tlv(a->u.ptr, BGP_AIGP_METRIC);
-
-  if (!b)
-    bsprintf(buf, "?");
-  else
-    bsprintf(buf, "%lu", get_u64(b + 3));
-}
-
-
 static void
 bgp_export_large_community(struct bgp_export_state *s, eattr *a)
 {
@@ -905,20 +633,20 @@ bgp_export_mpls_label_stack(struct bgp_export_state *s, eattr *a)
 
   /* Perhaps we should just ignore it? */
   if (!s->mpls)
-    REJECT("Unexpected MPLS stack");
+    WITHDRAW("Unexpected MPLS stack");
 
   /* Empty MPLS stack is not allowed */
   if (!lnum)
-    REJECT("Malformed MPLS stack - empty");
+    WITHDRAW("Malformed MPLS stack - empty");
 
   /* This is ugly, but we must ensure that labels fit into NLRI field */
   if ((24*lnum + (net_is_vpn(n) ? 64 : 0) + net_pxlen(n)) > 255)
-    REJECT("Malformed MPLS stack - too many labels (%u)", lnum);
+    WITHDRAW("Malformed MPLS stack - too many labels (%u)", lnum);
 
   for (uint i = 0; i < lnum; i++)
   {
     if (labels[i] > 0xfffff)
-      REJECT("Malformed MPLS stack - invalid label (%u)", labels[i]);
+      WITHDRAW("Malformed MPLS stack - invalid label (%u)", labels[i]);
 
     /* TODO: Check for special-purpose label values? */
   }
@@ -943,7 +671,7 @@ bgp_decode_mpls_label_stack(struct bgp_parse_state *s, uint code UNUSED, uint fl
 }
 
 static void
-bgp_format_mpls_label_stack(const eattr *a, byte *buf, uint size)
+bgp_format_mpls_label_stack(eattr *a, byte *buf, uint size)
 {
   u32 *labels = (u32 *) a->u.ptr->data;
   uint lnum = a->u.ptr->length / 4;
@@ -1092,15 +820,6 @@ static const struct bgp_attr_desc bgp_attr_table[] = {
     .decode = bgp_decode_as4_aggregator,
     .format = bgp_format_aggregator,
   },
-  [BA_AIGP] = {
-    .name = "aigp",
-    .type = EAF_TYPE_OPAQUE,
-    .flags = BAF_OPTIONAL | BAF_DECODE_FLAGS,
-    .export = bgp_export_aigp,
-    .encode = bgp_encode_raw,
-    .decode = bgp_decode_aigp,
-    .format = bgp_format_aigp,
-  },
   [BA_LARGE_COMMUNITY] = {
     .name = "large_community",
     .type = EAF_TYPE_LC_SET,
@@ -1199,7 +918,7 @@ bgp_export_attrs(struct bgp_export_state *s, ea_list *attrs)
   for (i = 0; i < count; i++)
     bgp_export_attr(s, &new->attrs[i], new);
 
-  if (s->err_reject)
+  if (s->err_withdraw)
     return NULL;
 
   return new;
@@ -1302,8 +1021,7 @@ bgp_decode_attr(struct bgp_parse_state *s, uint code, uint flags, byte *data, ui
     const struct bgp_attr_desc *desc = &bgp_attr_table[code];
 
     /* Handle conflicting flags; RFC 7606 3 (c) */
-    if (((flags ^ desc->flags) & (BAF_OPTIONAL | BAF_TRANSITIVE)) &&
-	!(desc->flags & BAF_DECODE_FLAGS))
+    if ((flags ^ desc->flags) & (BAF_OPTIONAL | BAF_TRANSITIVE))
       WITHDRAW("Malformed %s attribute - conflicting flags (%02x)", desc->name, flags);
 
     desc->decode(s, code, flags, data, len, to);
@@ -1397,19 +1115,19 @@ bgp_decode_attrs(struct bgp_parse_state *s, byte *data, uint len)
 
   /* Reject routes with our ASN in AS_PATH attribute */
   if (bgp_as_path_loopy(p, attrs, p->local_as))
-    goto loop;
+    goto withdraw;
 
   /* Reject routes with our Confederation ID in AS_PATH attribute; RFC 5065 4.0 */
   if ((p->public_as != p->local_as) && bgp_as_path_loopy(p, attrs, p->public_as))
-    goto loop;
+    goto withdraw;
 
   /* Reject routes with our Router ID in ORIGINATOR_ID attribute; RFC 4456 8 */
   if (p->is_internal && bgp_originator_id_loopy(p, attrs))
-    goto loop;
+    goto withdraw;
 
   /* Reject routes with our Cluster ID in CLUSTER_LIST attribute; RFC 4456 8 */
   if (p->rr_client && bgp_cluster_list_loopy(p, attrs))
-    goto loop;
+    goto withdraw;
 
   /* If there is no local preference, define one */
   if (!BIT32_TEST(s->attrs_seen, BA_LOCAL_PREF))
@@ -1430,21 +1148,6 @@ withdraw:
 
   s->err_withdraw = 1;
   return NULL;
-
-loop:
-  /* Loops are handled as withdraws, but ignored silently. Do not set err_withdraw. */
-  return NULL;
-}
-
-void
-bgp_finish_attrs(struct bgp_parse_state *s, rta *a)
-{
-  /* AIGP test here instead of in bgp_decode_aigp() - we need to know channel */
-  if (BIT32_TEST(s->attrs_seen, BA_AIGP) && !s->channel->cf->aigp)
-  {
-    REPORT("Discarding AIGP attribute received on non-AIGP session");
-    bgp_unset_attr(&a->eattrs, s->pool, BA_AIGP);
-  }
 }
 
 
@@ -1458,7 +1161,7 @@ bgp_finish_attrs(struct bgp_parse_state *s, rta *a)
 #define RBH_FN(a,h)		h
 
 #define RBH_REHASH		bgp_rbh_rehash
-#define RBH_PARAMS		/8, *2, 2, 2, 12, 20
+#define RBH_PARAMS		/8, *2, 2, 2, 8, 20
 
 
 HASH_DEFINE_REHASH_FN(RBH, struct bgp_bucket)
@@ -1515,7 +1218,6 @@ bgp_get_bucket(struct bgp_channel *c, ea_list *new)
 
   /* Create the bucket */
   b = mb_alloc(c->pool, size);
-  *b = (struct bgp_bucket) { };
   init_list(&b->prefixes);
   b->hash = hash;
 
@@ -1600,7 +1302,7 @@ bgp_withdraw_bucket(struct bgp_channel *c, struct bgp_bucket *b)
 #define PXH_FN(n,i,h)		h
 
 #define PXH_REHASH		bgp_pxh_rehash
-#define PXH_PARAMS		/8, *2, 2, 2, 12, 24
+#define PXH_PARAMS		/8, *2, 2, 2, 8, 24
 
 
 HASH_DEFINE_REHASH_FN(PXH, struct bgp_prefix)
@@ -1626,8 +1328,7 @@ bgp_free_prefix_table(struct bgp_channel *c)
 static struct bgp_prefix *
 bgp_get_prefix(struct bgp_channel *c, net_addr *net, u32 path_id)
 {
-  /* We must use a different hash function than the rtable */
-  u32 hash = u32_hash(net_hash(net) ^ u32_hash(path_id));
+  u32 hash = net_hash(net) ^ u32_hash(path_id);
   struct bgp_prefix *px = HASH_FIND(c->prefix_hash, PXH, net, path_id, hash);
 
   if (px)
@@ -1641,7 +1342,8 @@ bgp_get_prefix(struct bgp_channel *c, net_addr *net, u32 path_id)
   else
     px = mb_alloc(c->pool, sizeof(struct bgp_prefix) + net->length);
 
-  *px = (struct bgp_prefix) { };
+  px->buck_node.next = NULL;
+  px->buck_node.prev = NULL;
   px->hash = hash;
   px->path_id = path_id;
   net_copy(px->net, net);
@@ -1684,10 +1386,6 @@ bgp_preexport(struct proto *P, rte **new, struct linpool *pool UNUSED)
   if (src == NULL)
     return 0;
 
-  /* Reject flowspec that failed validation */
-  if ((e->attrs->dest == RTD_UNREACHABLE) && net_is_flow(e->net->n.addr))
-      return -1;
-
   /* IBGP route reflection, RFC 4456 */
   if (p->is_internal && src->is_internal && (p->local_as == src->local_as))
   {
@@ -1783,16 +1481,6 @@ bgp_update_attrs(struct bgp_proto *p, struct bgp_channel *c, rte *e, ea_list *at
   if (p->is_interior && ! bgp_find_attr(attrs0, BA_LOCAL_PREF))
     bgp_set_attr_u32(&attrs, pool, BA_LOCAL_PREF, 0, p->cf->default_local_pref);
 
-  /* AIGP attribute - accumulate local metric or originate new one */
-  u64 metric;
-  if (s.local_next_hop &&
-      (bgp_total_aigp_metric_(e, &metric, &ad) ||
-       (c->cf->aigp_originate && bgp_init_aigp_metric(e, &metric, &ad))))
-  {
-    ad = bgp_aigp_set_metric(pool, ad, metric);
-    bgp_set_attr_ptr(&attrs, pool, BA_AIGP, 0, ad);
-  }
-
   /* IBGP route reflection, RFC 4456 */
   if (src && src->is_internal && p->is_internal && (src->local_as == p->local_as))
   {
@@ -1857,10 +1545,6 @@ bgp_rt_notify(struct proto *P, struct channel *C, net *n, rte *new, rte *old)
   {
     struct ea_list *attrs = bgp_update_attrs(p, c, new, new->attrs->eattrs, bgp_linpool2);
 
-    /* Error during attribute processing */
-    if (!attrs)
-      log(L_ERR "%s: Invalid route %N withdrawn", p->p.name, n->n.addr);
-
     /* If attributes are invalid, we fail back to withdraw */
     buck = attrs ? bgp_get_bucket(c, attrs) : bgp_get_withdraw_bucket(c);
     path = new->attrs->src->global_id;
@@ -1894,6 +1578,12 @@ bgp_get_neighbor(rte *r)
   return p->cf->confederation ?: p->local_as;
 }
 
+static inline int
+rte_resolvable(rte *rt)
+{
+  return rt->attrs->dest == RTD_UNICAST;
+}
+
 static inline int
 rte_stale(rte *r)
 {
@@ -1949,14 +1639,6 @@ bgp_rte_better(rte *new, rte *old)
   if (n < o)
     return 0;
 
-  /* RFC 7311 4.1 - Apply AIGP metric */
-  u64 n2 = bgp_total_aigp_metric(new);
-  u64 o2 = bgp_total_aigp_metric(old);
-  if (n2 < o2)
-    return 1;
-  if (n2 > o2)
-    return 0;
-
   /* RFC 4271 9.1.2.2. a)  Use AS path lengths */
   if (new_bgp->cf->compare_path_lengths || old_bgp->cf->compare_path_lengths)
   {
@@ -2065,7 +1747,7 @@ bgp_rte_mergable(rte *pri, rte *sec)
     return 0;
 
   /* RFC 4271 9.1.2.1. Route resolvability test */
-  if (rte_resolvable(pri) != rte_resolvable(sec))
+  if (!rte_resolvable(sec))
     return 0;
 
   /* LLGR draft - depreference stale routes */
@@ -2151,7 +1833,7 @@ bgp_rte_recalculate(rtable *table, net *net, rte *new, rte *old, rte *old_best)
   rte *key = new ? new : old;
   u32 lpref = key->pref;
   u32 lasn = bgp_get_neighbor(key);
-  int old_suppressed = old ? old->u.bgp.suppressed : 0;
+  int old_is_group_best = 0;
 
   /*
    * Proper RFC 4271 path selection is a bit complicated, it cannot be
@@ -2198,7 +1880,7 @@ bgp_rte_recalculate(rtable *table, net *net, rte *new, rte *old, rte *old_best)
    * We could find the best-in-group and then make some shortcuts like
    * in rte_recalculate, but as we would have to walk through all
    * net->routes just to find it, it is probably not worth. So we
-   * just have one simple fast case that use just the old route.
+   * just have two simpler fast cases that use just the old route.
    * We also set suppressed flag to avoid using it in bgp_rte_better().
    */
 
@@ -2207,11 +1889,23 @@ bgp_rte_recalculate(rtable *table, net *net, rte *new, rte *old, rte *old_best)
 
   if (old)
   {
+    old_is_group_best = !old->u.bgp.suppressed;
     old->u.bgp.suppressed = 1;
+    int new_is_better = new && bgp_rte_better(new, old);
 
-    /* The fast case - replace not best with worse (or remove not best) */
-    if (old_suppressed && !(new && bgp_rte_better(new, old)))
+    /* The first case - replace not best with worse (or remove not best) */
+    if (!old_is_group_best && !new_is_better)
       return 0;
+
+    /* The second case - replace the best with better */
+    if (old_is_group_best && new_is_better)
+    {
+      /* new is best-in-group, the see discussion below - this is
+	 a special variant of NBG && OBG. From OBG we can deduce
+	 that same_group(old_best) iff (old == old_best)  */
+      new->u.bgp.suppressed = 0;
+      return (old == old_best);
+    }
   }
 
   /* The default case - find a new best-in-group route */
@@ -2228,16 +1922,6 @@ bgp_rte_recalculate(rtable *table, net *net, rte *new, rte *old, rte *old_best)
   if (!r)
     return 0;
 
-  /* Found if new is mergable with best-in-group */
-  if (new && (new != r) && bgp_rte_mergable(r, new))
-    new->u.bgp.suppressed = 0;
-
-  /* Found all existing routes mergable with best-in-group */
-  for (s=net->routes; rte_is_valid(s); s=s->next)
-    if (use_deterministic_med(s) && same_group(s, lpref, lasn))
-      if ((s != r) && bgp_rte_mergable(r, s))
-	s->u.bgp.suppressed = 0;
-
   /* Found best-in-group */
   r->u.bgp.suppressed = 0;
 
@@ -2251,9 +1935,9 @@ bgp_rte_recalculate(rtable *table, net *net, rte *new, rte *old, rte *old_best)
    * rte_recalculate() without ignore that possibility).
    *
    * There are three possible cases according to whether the old route
-   * was the best in group (OBG, i.e. !old_suppressed) and whether the
-   * new route is the best in group (NBG, tested by r == new). These
-   * cases work even if old or new is NULL.
+   * was the best in group (OBG, stored in old_is_group_best) and
+   * whether the new route is the best in group (NBG, tested by r == new).
+   * These cases work even if old or new is NULL.
    *
    * NBG -> new is a possible candidate for the best route, so we just
    *        check for the first reason using same_group().
@@ -2268,7 +1952,7 @@ bgp_rte_recalculate(rtable *table, net *net, rte *new, rte *old, rte *old_best)
   if (r == new)
     return old_best && same_group(old_best, lpref, lasn);
   else
-    return !old_suppressed;
+    return old_is_group_best;
 }
 
 struct rte *
@@ -2340,7 +2024,7 @@ bgp_process_as4_attrs(ea_list **attrs, struct linpool *pool)
 }
 
 int
-bgp_get_attr(const eattr *a, byte *buf, int buflen)
+bgp_get_attr(eattr *a, byte *buf, int buflen)
 {
   uint i = EA_ID(a->id);
   const struct bgp_attr_desc *d;
@@ -2380,12 +2064,7 @@ bgp_get_route_info(rte *e, byte *buf)
   if (rte_stale(e))
     buf += bsprintf(buf, "s");
 
-  u64 metric = bgp_total_aigp_metric(e);
-  if (metric < BGP_AIGP_MAX)
-  {
-    buf += bsprintf(buf, "/%lu", metric);
-  }
-  else if (e->attrs->igp_metric)
+  if (e->attrs->hostentry)
   {
     if (!rte_resolvable(e))
       buf += bsprintf(buf, "/-");
