@@ -185,20 +185,14 @@ bgp_find_af_caps(struct bgp_caps *caps, u32 afi)
 }
 
 static struct bgp_af_caps *
-bgp_get_af_caps(struct bgp_caps **pcaps, u32 afi)
+bgp_get_af_caps(struct bgp_caps *caps, u32 afi)
 {
-  struct bgp_caps *caps = *pcaps;
   struct bgp_af_caps *ac;
 
   WALK_AF_CAPS(caps, ac)
     if (ac->afi == afi)
       return ac;
 
-  uint n = caps->af_count;
-  if (uint_is_pow2(n))
-    *pcaps = caps = mb_realloc(caps, sizeof(struct bgp_caps) +
-			       (2 * n) * sizeof(struct bgp_af_caps));
-
   ac = &caps->af_data[caps->af_count++];
   memset(ac, 0, sizeof(struct bgp_af_caps));
   ac->afi = afi;
@@ -238,7 +232,6 @@ bgp_prepare_capabilities(struct bgp_conn *conn)
   caps->ext_messages = p->cf->enable_extended_messages;
   caps->route_refresh = p->cf->enable_refresh;
   caps->enhanced_refresh = p->cf->enable_refresh;
-  caps->role = p->cf->local_role;
 
   if (caps->as4_support)
     caps->as4_number = p->public_as;
@@ -253,16 +246,8 @@ bgp_prepare_capabilities(struct bgp_conn *conn)
   if (p->cf->llgr_mode)
     caps->llgr_aware = 1;
 
-  if (p->cf->enable_hostname && config->hostname)
-  {
-    size_t length = strlen(config->hostname);
-    char *hostname = mb_allocz(p->p.pool, length+1);
-    memcpy(hostname, config->hostname, length+1);
-    caps->hostname = hostname;
-  }
-
   /* Allocate and fill per-AF fields */
-  BGP_WALK_CHANNELS(p, c)
+  WALK_LIST(c, p->p.channels)
   {
     ac = &caps->af_data[caps->af_count++];
     ac->afi = c->afi;
@@ -309,8 +294,8 @@ bgp_write_capabilities(struct bgp_conn *conn, byte *buf)
 
   /*
    * Note that max length is ~ 22+21*af_count. With max 12 channels that is
-   * 274. We are limited just by buffer size (4096, minus header), as we support
-   * extended optional parameres. Therefore, we have enough space for expansion.
+   * 274. Option limit is 253 and buffer size is 4096, so we cannot overflow
+   * unless we add new capabilities or more AFs. XXXXX
    */
 
   WALK_AF_CAPS(caps, ac)
@@ -351,13 +336,6 @@ bgp_write_capabilities(struct bgp_conn *conn, byte *buf)
     *buf++ = 0;			/* Capability data length */
   }
 
-  if (caps->role != BGP_ROLE_UNDEFINED)
-  {
-    *buf++ = 9;			/* Capability 9: Announce chosen BGP role */
-    *buf++ = 1;			/* Capability data length */
-    *buf++ = caps->role;
-  }
-
   if (caps->gr_aware)
   {
     *buf++ = 64;		/* Capability 64: Support for graceful restart */
@@ -428,50 +406,19 @@ bgp_write_capabilities(struct bgp_conn *conn, byte *buf)
     data[-1] = buf - data;
   }
 
-  if (caps->hostname)
-  {
-    *buf++ = 73;                /* Capability 73: Hostname */
-    *buf++ = 0;			/* Capability data length */
-    data = buf;
-
-    /* Hostname */
-    size_t length = strlen(caps->hostname);
-    *buf++ = length;
-    memcpy(buf, caps->hostname, length);
-    buf += length;
-
-    /* Domain, not implemented */
-    *buf++ = 0;
-
-    data[-1] = buf - data;
-  }
-
   caps->length = buf - buf_head;
 
   return buf;
 }
 
-static int
-bgp_read_capabilities(struct bgp_conn *conn, byte *pos, int len)
+static void
+bgp_read_capabilities(struct bgp_conn *conn, struct bgp_caps *caps, byte *pos, int len)
 {
   struct bgp_proto *p = conn->bgp;
-  struct bgp_caps *caps;
   struct bgp_af_caps *ac;
-  uint err_subcode = 0;
   int i, cl;
   u32 af;
 
-  if (!conn->remote_caps)
-  {
-    caps = mb_allocz(p->p.pool, sizeof(struct bgp_caps) + sizeof(struct bgp_af_caps));
-    caps->role = BGP_ROLE_UNDEFINED;
-  }
-  else
-  {
-    caps = conn->remote_caps;
-    conn->remote_caps = NULL;
-  }
-
   caps->length += len;
 
   while (len > 0)
@@ -490,7 +437,7 @@ bgp_read_capabilities(struct bgp_conn *conn, byte *pos, int len)
 	goto err;
 
       af = get_af4(pos+2);
-      ac = bgp_get_af_caps(&caps, af);
+      ac = bgp_get_af_caps(caps, af);
       ac->ready = 1;
       break;
 
@@ -513,7 +460,7 @@ bgp_read_capabilities(struct bgp_conn *conn, byte *pos, int len)
 	  continue;
 
 	af = get_af4(pos+2+i);
-	ac = bgp_get_af_caps(&caps, af);
+	ac = bgp_get_af_caps(caps, af);
 	ac->ext_next_hop = 1;
       }
       break;
@@ -525,21 +472,6 @@ bgp_read_capabilities(struct bgp_conn *conn, byte *pos, int len)
       caps->ext_messages = 1;
       break;
 
-    case  9: /* BGP role capability, RFC 9234 */
-      if (cl != 1)
-        goto err;
-
-      /* Reserved value */
-      if (pos[2] == BGP_ROLE_UNDEFINED)
-      { err_subcode = 11; goto err; }
-
-      /* Multiple inconsistent values */
-      if ((caps->role != BGP_ROLE_UNDEFINED) && (caps->role != pos[2]))
-      { err_subcode = 11; goto err; }
-
-      caps->role = pos[2];
-      break;
-
     case 64: /* Graceful restart capability, RFC 4724 */
       if (cl % 4 != 2)
 	goto err;
@@ -558,7 +490,7 @@ bgp_read_capabilities(struct bgp_conn *conn, byte *pos, int len)
       for (i = 2; i < cl; i += 4)
       {
 	af = get_af3(pos+2+i);
-	ac = bgp_get_af_caps(&caps, af);
+	ac = bgp_get_af_caps(caps, af);
 	ac->gr_able = 1;
 	ac->gr_af_flags = pos[2+i+3];
       }
@@ -590,7 +522,7 @@ bgp_read_capabilities(struct bgp_conn *conn, byte *pos, int len)
       for (i = 0; i < cl; i += 4)
       {
 	af = get_af3(pos+2+i);
-	ac = bgp_get_af_caps(&caps, af);
+	ac = bgp_get_af_caps(caps, af);
 	ac->add_path = pos[2+i+3];
       }
       break;
@@ -619,28 +551,13 @@ bgp_read_capabilities(struct bgp_conn *conn, byte *pos, int len)
       for (i = 0; i < cl; i += 7)
       {
 	af = get_af3(pos+2+i);
-	ac = bgp_get_af_caps(&caps, af);
+	ac = bgp_get_af_caps(caps, af);
 	ac->llgr_able = 1;
 	ac->llgr_flags = pos[2+i+3];
 	ac->llgr_time = get_u24(pos + 2+i+4);
       }
       break;
 
-    case 73: /* Hostname, RFC draft */
-      if ((cl < 2) || (cl < 2 + pos[2]))
-        goto err;
-
-      int length = pos[2];
-      char *hostname = mb_allocz(p->p.pool, length+1);
-      memcpy(hostname, pos + 3, length);
-      hostname[length] = 0;
-
-      for (i = 0; i < length; i++)
-        if (hostname[i] < ' ')
-          hostname[i] = ' ';
-
-      caps->hostname = hostname;
-
       /* We can safely ignore all other capabilities */
     }
 
@@ -660,13 +577,11 @@ bgp_read_capabilities(struct bgp_conn *conn, byte *pos, int len)
     }
   }
 
-  conn->remote_caps = caps;
-  return 0;
+  return;
 
 err:
-  mb_free(caps);
-  bgp_error(conn, 2, err_subcode, NULL, 0);
-  return -1;
+  bgp_error(conn, 2, 0, NULL, 0);
+  return;
 }
 
 static int
@@ -681,7 +596,7 @@ bgp_check_capabilities(struct bgp_conn *conn)
   /* This is partially overlapping with bgp_conn_enter_established_state(),
      but we need to run this just after we receive OPEN message */
 
-  BGP_WALK_CHANNELS(p, c)
+  WALK_LIST(c, p->p.channels)
   {
     const struct bgp_af_caps *loc = bgp_find_af_caps(local,  c->afi);
     const struct bgp_af_caps *rem = bgp_find_af_caps(remote, c->afi);
@@ -706,68 +621,43 @@ bgp_check_capabilities(struct bgp_conn *conn)
 }
 
 static int
-bgp_read_options(struct bgp_conn *conn, byte *pos, uint len, uint rest)
+bgp_read_options(struct bgp_conn *conn, byte *pos, int len)
 {
   struct bgp_proto *p = conn->bgp;
-  int ext = 0;
+  struct bgp_caps *caps;
+  int ol;
 
-  /* Handle extended length (draft-ietf-idr-ext-opt-param-07) */
-  if ((len > 0) && (rest > 0) && (pos[0] == 255))
-  {
-    if (rest < 3)
-      goto err;
-
-    /* Update pos/len to describe optional data */
-    len = get_u16(pos+1);
-    ext = 1;
-    pos += 3;
-    rest -= 3;
-  }
-
-  /* Verify that optional data fits into OPEN packet */
-  if (len > rest)
-    goto err;
-
-  /* Length of option parameter header */
-  uint hlen = ext ? 3 : 2;
+  /* Max number of announced AFIs is limited by max option length (255) */
+  caps = alloca(sizeof(struct bgp_caps) + 64 * sizeof(struct bgp_af_caps));
+  memset(caps, 0, sizeof(struct bgp_caps));
 
   while (len > 0)
   {
-    if (len < hlen)
-      goto err;
+    if ((len < 2) || (len < (2 + pos[1])))
+    { bgp_error(conn, 2, 0, NULL, 0); return -1; }
 
-    uint otype = get_u8(pos);
-    uint olen = ext ? get_u16(pos+1) : get_u8(pos+1);
-
-    if (len < (hlen + olen))
-      goto err;
-
-    if (otype == 2)
+    ol = pos[1];
+    if (pos[0] == 2)
     {
       /* BGP capabilities, RFC 5492 */
       if (p->cf->capabilities)
-	if (bgp_read_capabilities(conn, pos + hlen, olen) < 0)
-	  return -1;
+	bgp_read_capabilities(conn, caps, pos + 2, ol);
     }
     else
     {
       /* Unknown option */
-      bgp_error(conn, 2, 4, pos, hlen + olen);
+      bgp_error(conn, 2, 4, pos, ol); /* FIXME: ol or ol+2 ? */
       return -1;
     }
 
-    ADVANCE(pos, len, hlen + olen);
+    ADVANCE(pos, len, 2 + ol);
   }
 
-  /* Prepare empty caps if no capability option was announced */
-  if (!conn->remote_caps)
-    conn->remote_caps = mb_allocz(p->p.pool, sizeof(struct bgp_caps));
+  uint n = sizeof(struct bgp_caps) + caps->af_count * sizeof(struct bgp_af_caps);
+  conn->remote_caps = mb_allocz(p->p.pool, n);
+  memcpy(conn->remote_caps, caps, n);
 
   return 0;
-
-err:
-  bgp_error(conn, 2, 0, NULL, 0);
-  return -1;
 }
 
 static byte *
@@ -786,29 +676,12 @@ bgp_create_open(struct bgp_conn *conn, byte *buf)
   if (p->cf->capabilities)
   {
     /* Prepare local_caps and write capabilities to buffer */
-    byte *pos = buf+12;
-    byte *end = bgp_write_capabilities(conn, pos);
-    uint len = end - pos;
+    byte *end = bgp_write_capabilities(conn, buf+12);
+    uint len = end - (buf+12);
 
-    if (len < 254)
-    {
-      buf[9] = len + 2;		/* Optional parameters length */
-      buf[10] = 2;		/* Option 2: Capability list */
-      buf[11] = len;		/* Option data length */
-    }
-    else /* draft-ietf-idr-ext-opt-param-07 */
-    {
-      /* Move capabilities 4 B forward */
-      memmove(buf + 16, pos, len);
-      pos = buf + 16;
-      end = pos + len;
-
-      buf[9] = 255;		/* Non-ext OP length, fake */
-      buf[10] = 255;		/* Non-ext OP type, signals extended length */
-      put_u16(buf+11, len + 3);	/* Extended optional parameters length */
-      buf[13] = 2;		/* Option 2: Capability list */
-      put_u16(buf+14, len);	/* Option extended data length */
-    }
+    buf[9] = len + 2;		/* Optional parameters length */
+    buf[10] = 2;		/* Option 2: Capability list */
+    buf[11] = len;		/* Option data length */
 
     return end;
   }
@@ -832,8 +705,8 @@ bgp_rx_open(struct bgp_conn *conn, byte *pkt, uint len)
   if (conn->state != BS_OPENSENT)
   { bgp_error(conn, 5, fsm_err_subcode[conn->state], NULL, 0); return; }
 
-  /* Check message length */
-  if (len < 29)
+  /* Check message contents */
+  if (len < 29 || len != 29 + (uint) pkt[28])
   { bgp_error(conn, 1, 2, pkt+16, 2); return; }
 
   if (pkt[19] != BGP_VERSION)
@@ -844,28 +717,12 @@ bgp_rx_open(struct bgp_conn *conn, byte *pkt, uint len)
   id = get_u32(pkt+24);
   BGP_TRACE(D_PACKETS, "Got OPEN(as=%d,hold=%d,id=%R)", asn, hold, id);
 
-  if (bgp_read_options(conn, pkt+29, pkt[28], len-29) < 0)
+  if (bgp_read_options(conn, pkt+29, pkt[28]) < 0)
     return;
 
-  /* RFC 4271 4.2 - hold time must be either 0 or at least 3 */
   if (hold > 0 && hold < 3)
   { bgp_error(conn, 2, 6, pkt+22, 2); return; }
 
-  /* Compute effective hold and keepalive times */
-  uint hold_time = MIN(hold, p->cf->hold_time);
-  uint keepalive_time = p->cf->keepalive_time ?
-    (p->cf->keepalive_time * hold_time / p->cf->hold_time) :
-    hold_time / 3;
-
-  /* Keepalive time might be rounded down to zero */
-  if (hold_time && !keepalive_time)
-    keepalive_time = 1;
-
-  /* Check effective values against configured minimums */
-  if ((hold_time < p->cf->min_hold_time) ||
-      (keepalive_time < p->cf->min_keepalive_time))
-  { bgp_error(conn, 2, 6, pkt+22, 2); return; }
-
   /* RFC 6286 2.2 - router ID is nonzero and AS-wide unique */
   if (!id || (p->is_internal && id == p->local_id))
   { bgp_error(conn, 2, 3, pkt+24, -4); return; }
@@ -897,22 +754,6 @@ bgp_rx_open(struct bgp_conn *conn, byte *pkt, uint len)
     conn->received_as = asn;
   }
 
-  /* RFC 9234 4.2 - check role agreement */
-  u8 local_role = p->cf->local_role;
-  u8 neigh_role = caps->role;
-
-  if ((local_role != BGP_ROLE_UNDEFINED) &&
-      (neigh_role != BGP_ROLE_UNDEFINED) &&
-      !((local_role == BGP_ROLE_PEER && neigh_role == BGP_ROLE_PEER) ||
-	(local_role == BGP_ROLE_CUSTOMER && neigh_role == BGP_ROLE_PROVIDER) ||
-	(local_role == BGP_ROLE_PROVIDER && neigh_role == BGP_ROLE_CUSTOMER) ||
-	(local_role == BGP_ROLE_RS_CLIENT && neigh_role == BGP_ROLE_RS_SERVER) ||
-	(local_role == BGP_ROLE_RS_SERVER && neigh_role == BGP_ROLE_RS_CLIENT)))
-  { bgp_error(conn, 2, 11, NULL, 0); return; }
-
-  if ((p->cf->require_roles) && (neigh_role == BGP_ROLE_UNDEFINED))
-  { bgp_error(conn, 2, 11, NULL, 0); return; }
-
   /* Check the other connection */
   other = (conn == &p->outgoing_conn) ? &p->incoming_conn : &p->outgoing_conn;
   switch (other->state)
@@ -963,8 +804,8 @@ bgp_rx_open(struct bgp_conn *conn, byte *pkt, uint len)
   }
 
   /* Update our local variables */
-  conn->hold_time = hold_time;
-  conn->keepalive_time = keepalive_time;
+  conn->hold_time = MIN(hold, p->cf->hold_time);
+  conn->keepalive_time = p->cf->keepalive_time ? : conn->hold_time / 3;
   conn->as4_session = conn->local_caps->as4_support && caps->as4_support;
   conn->ext_messages = conn->local_caps->ext_messages && caps->ext_messages;
   p->remote_id = id;
@@ -991,15 +832,11 @@ bgp_rx_open(struct bgp_conn *conn, byte *pkt, uint len)
 #define WITHDRAW(msg, args...) \
   ({ REPORT(msg, ## args); s->err_withdraw = 1; return; })
 
-#define REJECT(msg, args...)						\
-  ({ log(L_ERR "%s: " msg, s->proto->p.name, ## args); s->err_reject = 1; return; })
-
 #define BAD_AFI		"Unexpected AF <%u/%u> in UPDATE"
 #define BAD_NEXT_HOP	"Invalid NEXT_HOP attribute"
 #define NO_NEXT_HOP	"Missing NEXT_HOP attribute"
 #define NO_LABEL_STACK	"Missing MPLS stack"
 
-#define MISMATCHED_AF	" - mismatched address family (%I for %s)"
 
 static void
 bgp_apply_next_hop(struct bgp_parse_state *s, rta *a, ip_addr gw, ip_addr ll)
@@ -1012,32 +849,25 @@ bgp_apply_next_hop(struct bgp_parse_state *s, rta *a, ip_addr gw, ip_addr ll)
     neighbor *nbr = NULL;
 
     /* GW_DIRECT -> single_hop -> p->neigh != NULL */
-    if (ipa_nonzero2(gw))
+    if (ipa_nonzero(gw))
       nbr = neigh_find(&p->p, gw, NULL, 0);
     else if (ipa_nonzero(ll))
       nbr = neigh_find(&p->p, ll, p->neigh->iface, 0);
-    else
-      WITHDRAW(BAD_NEXT_HOP " - zero address");
 
-    if (!nbr)
-      WITHDRAW(BAD_NEXT_HOP " - address %I not directly reachable", ipa_nonzero(gw) ? gw : ll);
-
-    if (nbr->scope == SCOPE_HOST)
-      WITHDRAW(BAD_NEXT_HOP " - address %I is local", nbr->addr);
+    if (!nbr || (nbr->scope == SCOPE_HOST))
+      WITHDRAW(BAD_NEXT_HOP);
 
     a->dest = RTD_UNICAST;
     a->nh.gw = nbr->addr;
     a->nh.iface = nbr->iface;
-    a->igp_metric = c->cf->cost;
   }
   else /* GW_RECURSIVE */
   {
-    if (ipa_zero2(gw))
-      WITHDRAW(BAD_NEXT_HOP " - zero address");
+    if (ipa_zero(gw))
+      WITHDRAW(BAD_NEXT_HOP);
 
     rtable *tab = ipa_is_ip4(gw) ? c->igp_table_ip4 : c->igp_table_ip6;
-    ip_addr lla = (c->cf->next_hop_prefer == NHP_LOCAL) ? ll : IPA_NONE;
-    s->hostentry = rt_get_hostentry(tab, gw, lla, c->c.table);
+    s->hostentry = rt_get_hostentry(tab, gw, ll, c->c.table);
 
     if (!s->mpls)
       rta_apply_hostentry(a, s->hostentry, NULL);
@@ -1078,20 +908,6 @@ bgp_apply_mpls_labels(struct bgp_parse_state *s, rta *a, u32 *labels, uint lnum)
   }
 }
 
-static void
-bgp_apply_flow_validation(struct bgp_parse_state *s, const net_addr *n, rta *a)
-{
-  struct bgp_channel *c = s->channel;
-  int valid = rt_flowspec_check(c->base_table, c->c.table, n, a, s->proto->is_interior);
-  a->dest = valid ? RTD_NONE : RTD_UNREACHABLE;
-
-  /* Invalidate cached rta if dest changes */
-  if (s->cached_rta && (s->cached_rta->dest != a->dest))
-  {
-    rta_free(s->cached_rta);
-    s->cached_rta = NULL;
-  }
-}
 
 static int
 bgp_match_src(struct bgp_export_state *s, int mode)
@@ -1122,24 +938,19 @@ bgp_use_next_hop(struct bgp_export_state *s, eattr *a)
     return 1;
 
   /* Keep it when explicitly set in export filter */
-  if (a->fresh)
+  if (a->type & EAF_FRESH)
     return 1;
 
   /* Check for non-matching AF */
   if ((ipa_is_ip4(*nh) != bgp_channel_is_ipv4(c)) && !c->ext_next_hop)
     return 0;
 
-  /* Do not pass NEXT_HOP between different VRFs */
-  if (p->p.vrf_set && s->src && s->src->p.vrf_set && (p->p.vrf != s->src->p.vrf))
-    return 0;
-
   /* Keep it when exported to internal peers */
   if (p->is_interior && ipa_nonzero(*nh))
     return 1;
 
   /* Keep it when forwarded between single-hop BGPs on the same iface */
-  struct iface *ifa = (s->src && s->src->neigh && (s->src->p.proto_state != PS_DOWN)) ?
-    s->src->neigh->iface : NULL;
+  struct iface *ifa = (s->src && s->src->neigh) ? s->src->neigh->iface : NULL;
   return p->neigh && (p->neigh->iface == ifa);
 }
 
@@ -1162,10 +973,6 @@ bgp_use_gateway(struct bgp_export_state *s)
   if ((ipa_is_ip4(ra->nh.gw) != bgp_channel_is_ipv4(c)) && !c->ext_next_hop)
     return 0;
 
-  /* Do not use gateway from different VRF */
-  if (p->p.vrf_set && ra->nh.iface && (p->p.vrf != ra->nh.iface->master))
-    return 0;
-
   /* Use it when exported to internal peers */
   if (p->is_interior)
     return 1;
@@ -1192,14 +999,11 @@ bgp_update_next_hop_ip(struct bgp_export_state *s, eattr *a, ea_list **to)
 	uint lnum = ra->nh.labels ? ra->nh.labels : 1;
 	bgp_set_attr_data(to, s->pool, BA_MPLS_LABEL_STACK, 0, labels, lnum * 4);
       }
-      else
-	bgp_unset_attr(to, s->pool, BA_MPLS_LABEL_STACK);
     }
     else
     {
       ip_addr nh[2] = { s->channel->next_hop_addr, s->channel->link_addr };
       bgp_set_attr_data(to, s->pool, BA_NEXT_HOP, 0, nh, ipa_nonzero(nh[1]) ? 32 : 16);
-      s->local_next_hop = 1;
 
       /* TODO: Use local MPLS assigned label */
       if (s->mpls)
@@ -1207,36 +1011,34 @@ bgp_update_next_hop_ip(struct bgp_export_state *s, eattr *a, ea_list **to)
 	u32 implicit_null = BGP_MPLS_NULL;
 	bgp_set_attr_data(to, s->pool, BA_MPLS_LABEL_STACK, 0, &implicit_null, 4);
       }
-      else
-	bgp_unset_attr(to, s->pool, BA_MPLS_LABEL_STACK);
     }
   }
 
   /* Check if next hop is valid */
   a = bgp_find_attr(*to, BA_NEXT_HOP);
   if (!a)
-    REJECT(NO_NEXT_HOP);
+    WITHDRAW(NO_NEXT_HOP);
 
   ip_addr *nh = (void *) a->u.ptr->data;
   ip_addr peer = s->proto->remote_ip;
   uint len = a->u.ptr->length;
 
   /* Forbid zero next hop */
-  if (ipa_zero2(nh[0]) && ((len != 32) || ipa_zero(nh[1])))
-    REJECT(BAD_NEXT_HOP " - zero address");
+  if (ipa_zero(nh[0]) && ((len != 32) || ipa_zero(nh[1])))
+    WITHDRAW(BAD_NEXT_HOP);
 
   /* Forbid next hop equal to neighbor IP */
   if (ipa_equal(peer, nh[0]) || ((len == 32) && ipa_equal(peer, nh[1])))
-    REJECT(BAD_NEXT_HOP " - neighbor address %I", peer);
+    WITHDRAW(BAD_NEXT_HOP);
 
   /* Forbid next hop with non-matching AF */
   if ((ipa_is_ip4(nh[0]) != bgp_channel_is_ipv4(s->channel)) &&
       !s->channel->ext_next_hop)
-    REJECT(BAD_NEXT_HOP MISMATCHED_AF, nh[0], s->channel->desc->name);
+    WITHDRAW(BAD_NEXT_HOP);
 
   /* Just check if MPLS stack */
   if (s->mpls && !bgp_find_attr(*to, BA_MPLS_LABEL_STACK))
-    REJECT(NO_LABEL_STACK);
+    WITHDRAW(NO_LABEL_STACK);
 }
 
 static uint
@@ -1294,10 +1096,7 @@ bgp_decode_next_hop_ip(struct bgp_parse_state *s, byte *data, uint len, rta *a)
     nh[0] = ipa_from_ip6(get_ip6(data));
     nh[1] = ipa_from_ip6(get_ip6(data+16));
 
-    if (ipa_is_link_local(nh[0]))
-    { nh[1] = nh[0]; nh[0] = IPA_NONE; }
-
-    if (ipa_is_ip4(nh[0]) || !ipa_is_link_local(nh[1]))
+    if (ipa_is_ip4(nh[0]) || !ip6_is_link_local(nh[1]))
       nh[1] = IPA_NONE;
   }
   else
@@ -1307,7 +1106,7 @@ bgp_decode_next_hop_ip(struct bgp_parse_state *s, byte *data, uint len, rta *a)
     ad->length = 16;
 
   if ((bgp_channel_is_ipv4(c) != ipa_is_ip4(nh[0])) && !c->ext_next_hop)
-    WITHDRAW(BAD_NEXT_HOP MISMATCHED_AF, nh[0], c->desc->name);
+    WITHDRAW(BAD_NEXT_HOP);
 
   // XXXX validate next hop
 
@@ -1388,7 +1187,7 @@ bgp_decode_next_hop_vpn(struct bgp_parse_state *s, byte *data, uint len, rta *a)
     bgp_parse_error(s, 9);
 
   if ((bgp_channel_is_ipv4(c) != ipa_is_ip4(nh[0])) && !c->ext_next_hop)
-    WITHDRAW(BAD_NEXT_HOP MISMATCHED_AF, nh[0], c->desc->name);
+    WITHDRAW(BAD_NEXT_HOP);
 
   // XXXX validate next hop
 
@@ -1430,7 +1229,7 @@ bgp_update_next_hop_none(struct bgp_export_state *s, eattr *a, ea_list **to)
  */
 
 static void
-bgp_rte_update(struct bgp_parse_state *s, const net_addr *n, u32 path_id, rta *a0)
+bgp_rte_update(struct bgp_parse_state *s, net_addr *n, u32 path_id, rta *a0)
 {
   if (path_id != s->last_id)
   {
@@ -1443,10 +1242,6 @@ bgp_rte_update(struct bgp_parse_state *s, const net_addr *n, u32 path_id, rta *a
 
   if (!a0)
   {
-    /* Route update was changed to withdraw */
-    if (s->err_withdraw && s->reach_nlri_step)
-      REPORT("Invalid route %N withdrawn", n);
-
     /* Route withdraw */
     rte_update3(&s->channel->c, n, NULL, s->last_src);
     return;
@@ -1455,6 +1250,8 @@ bgp_rte_update(struct bgp_parse_state *s, const net_addr *n, u32 path_id, rta *a
   /* Prepare cached route attributes */
   if (s->cached_rta == NULL)
   {
+    a0->src = s->last_src;
+
     /* Workaround for rta_lookup() breaking eattrs */
     ea_list *ea = a0->eattrs;
     s->cached_rta = rta_lookup(a0);
@@ -1462,9 +1259,11 @@ bgp_rte_update(struct bgp_parse_state *s, const net_addr *n, u32 path_id, rta *a
   }
 
   rta *a = rta_clone(s->cached_rta);
-  rte *e = rte_get_temp(a, s->last_src);
+  rte *e = rte_get_temp(a);
 
   e->pflags = 0;
+  e->u.bgp.suppressed = 0;
+  e->u.bgp.stale = -1;
   rte_update3(&s->channel->c, n, e, s->last_src);
 }
 
@@ -1504,7 +1303,7 @@ bgp_decode_mpls_labels(struct bgp_parse_state *s, byte **pos, uint *len, uint *p
 
     /* RFC 8277 2.4 - withdraw does not have variable-size MPLS stack but
        fixed-size 24-bit Compatibility field, which MUST be ignored */
-    if (!s->reach_nlri_step)
+    if (!a && !s->err_withdraw)
       return;
   }
   while (!(label & BGP_MPLS_BOS));
@@ -1964,25 +1763,25 @@ bgp_decode_nlri_flow4(struct bgp_parse_state *s, byte *pos, uint len, rta *a)
       bgp_parse_error(s, 1);
     }
 
-    ip4_addr px = IP4_NONE;
-    uint pxlen = 0;
+    if (data[0] != FLOW_TYPE_DST_PREFIX)
+    {
+      log(L_REMOTE "%s: No dst prefix at first pos", s->proto->p.name);
+      bgp_parse_error(s, 1);
+    }
 
     /* Decode dst prefix */
-    if (data[0] == FLOW_TYPE_DST_PREFIX)
-    {
-      px = flow_read_ip4_part(data);
-      pxlen = flow_read_pxlen(data);
-    }
+    ip4_addr px = IP4_NONE;
+    uint pxlen = data[1];
+
+    // FIXME: Use some generic function
+    memcpy(&px, data+2, BYTES(pxlen));
+    px = ip4_and(ip4_ntoh(px), ip4_mkmask(pxlen));
 
     /* Prepare the flow */
     net_addr *n = alloca(sizeof(struct net_addr_flow4) + flen);
     net_fill_flow4(n, px, pxlen, pos, flen);
     ADVANCE(pos, len, flen);
 
-    /* Apply validation procedure per RFC 8955 (6) */
-    if (a && s->channel->cf->validate)
-      bgp_apply_flow_validation(s, n, a);
-
     bgp_rte_update(s, n, path_id, a);
   }
 }
@@ -2056,25 +1855,25 @@ bgp_decode_nlri_flow6(struct bgp_parse_state *s, byte *pos, uint len, rta *a)
       bgp_parse_error(s, 1);
     }
 
-    ip6_addr px = IP6_NONE;
-    uint pxlen = 0;
+    if (data[0] != FLOW_TYPE_DST_PREFIX)
+    {
+      log(L_REMOTE "%s: No dst prefix at first pos", s->proto->p.name);
+      bgp_parse_error(s, 1);
+    }
 
     /* Decode dst prefix */
-    if (data[0] == FLOW_TYPE_DST_PREFIX)
-    {
-      px = flow_read_ip6_part(data);
-      pxlen = flow_read_pxlen(data);
-    }
+    ip6_addr px = IP6_NONE;
+    uint pxlen = data[1];
+
+    // FIXME: Use some generic function
+    memcpy(&px, data+2, BYTES(pxlen));
+    px = ip6_and(ip6_ntoh(px), ip6_mkmask(pxlen));
 
     /* Prepare the flow */
     net_addr *n = alloca(sizeof(struct net_addr_flow6) + flen);
     net_fill_flow6(n, px, pxlen, pos, flen);
     ADVANCE(pos, len, flen);
 
-    /* Apply validation procedure per RFC 8955 (6) */
-    if (a && s->channel->cf->validate)
-      bgp_apply_flow_validation(s, n, a);
-
     bgp_rte_update(s, n, path_id, a);
   }
 }
@@ -2395,14 +2194,11 @@ bgp_create_update(struct bgp_channel *c, byte *buf)
 
 again: ;
 
-  struct lp_state tmpp;
-  lp_save(tmp_linpool, &tmpp);
-
   /* Initialize write state */
   struct bgp_write_state s = {
     .proto = p,
     .channel = c,
-    .pool = tmp_linpool,
+    .pool = bgp_linpool,
     .mp_reach = (c->afi != BGP_AF_IPV4) || c->ext_next_hop,
     .as4_session = p->as4_session,
     .add_path = c->add_path_tx,
@@ -2428,7 +2224,6 @@ again: ;
     if (EMPTY_LIST(buck->prefixes))
     {
       bgp_free_bucket(c, buck);
-      lp_restore(tmp_linpool, &tmpp);
       goto again;
     }
 
@@ -2442,10 +2237,7 @@ again: ;
       bgp_defer_bucket(c, buck);
 
     if (!res)
-    {
-      lp_restore(tmp_linpool, &tmpp);
       goto again;
-    }
 
     goto done;
   }
@@ -2455,8 +2247,7 @@ again: ;
 
 done:
   BGP_TRACE_RL(&rl_snd_update, D_PACKETS, "Sending UPDATE");
-  p->stats.tx_updates++;
-  lp_restore(tmp_linpool, &tmpp);
+  lp_flush(s.pool);
 
   return res;
 }
@@ -2491,7 +2282,6 @@ bgp_create_end_mark(struct bgp_channel *c, byte *buf)
   struct bgp_proto *p = (void *) c->c.proto;
 
   BGP_TRACE(D_PACKETS, "Sending END-OF-RIB");
-  p->stats.tx_updates++;
 
   return (c->afi == BGP_AF_IPV4) ?
     bgp_create_ip_end_mark(c, buf):
@@ -2550,10 +2340,8 @@ bgp_decode_nlri(struct bgp_parse_state *s, u32 afi, byte *nlri, uint len, ea_lis
     a->scope = SCOPE_UNIVERSE;
     a->from = s->proto->remote_ip;
     a->eattrs = ea;
-    a->pref = c->c.preference;
 
     c->desc->decode_next_hop(s, nh, nh_len, a);
-    bgp_finish_attrs(s, a);
 
     /* Handle withdraw during next hop decoding */
     if (s->err_withdraw)
@@ -2573,8 +2361,6 @@ bgp_rx_update(struct bgp_conn *conn, byte *pkt, uint len)
   ea_list *ea = NULL;
 
   BGP_TRACE_RL(&rl_rcv_update, D_PACKETS, "Got UPDATE");
-  p->last_rx_update = current_time();
-  p->stats.rx_updates++;
 
   /* Workaround for some BGP implementations that skip initial KEEPALIVE */
   if (conn->state == BS_OPENCONFIRM)
@@ -2585,13 +2371,10 @@ bgp_rx_update(struct bgp_conn *conn, byte *pkt, uint len)
 
   bgp_start_timer(conn->hold_timer, conn->hold_time);
 
-  struct lp_state tmpp;
-  lp_save(tmp_linpool, &tmpp);
-
   /* Initialize parse state */
   struct bgp_parse_state s = {
     .proto = p,
-    .pool = tmp_linpool,
+    .pool = bgp_linpool,
     .as4_session = p->as4_session,
   };
 
@@ -2657,8 +2440,6 @@ bgp_rx_update(struct bgp_conn *conn, byte *pkt, uint len)
   if (s.mp_unreach_len)
     bgp_decode_nlri(&s, s.mp_unreach_af, s.mp_unreach_nlri, s.mp_unreach_len, NULL, NULL, 0);
 
-  s.reach_nlri_step = 1;
-
   if (s.ip_reach_len)
     bgp_decode_nlri(&s, BGP_AF_IPV4, s.ip_reach_nlri, s.ip_reach_len,
 		    ea, s.ip_next_hop_data, s.ip_next_hop_len);
@@ -2669,7 +2450,7 @@ bgp_rx_update(struct bgp_conn *conn, byte *pkt, uint len)
 
 done:
   rta_free(s.cached_rta);
-  lp_restore(tmp_linpool, &tmpp);
+  lp_flush(s.pool);
   return;
 }
 
@@ -2867,9 +2648,6 @@ bgp_send(struct bgp_conn *conn, uint type, uint len)
   sock *sk = conn->sk;
   byte *buf = sk->tbuf;
 
-  conn->bgp->stats.tx_messages++;
-  conn->bgp->stats.tx_bytes += len;
-
   memset(buf, 0xff, 16);		/* Marker */
   put_u16(buf+16, len);
   buf[18] = type;
@@ -3057,7 +2835,6 @@ static struct {
   { 2, 6, "Unacceptable hold time" },
   { 2, 7, "Required capability missing" }, /* [RFC5492] */
   { 2, 8, "No supported AFI/SAFI" }, /* This error msg is nonstandard */
-  { 2,11, "Role mismatch" }, /* From Open Policy, RFC 9234 */
   { 3, 0, "Invalid UPDATE message" },
   { 3, 1, "Malformed attribute list" },
   { 3, 2, "Unrecognized well-known attribute" },
@@ -3123,7 +2900,7 @@ bgp_handle_message(struct bgp_proto *p, byte *data, uint len, byte **bp)
     return 1;
 
   /* Handle proper message */
-  if (msg_len + 1 > len)
+  if ((msg_len > 255) && (msg_len + 1 > len))
     return 0;
 
   /* Some elementary cleanup */
@@ -3139,7 +2916,7 @@ bgp_handle_message(struct bgp_proto *p, byte *data, uint len, byte **bp)
 void
 bgp_log_error(struct bgp_proto *p, u8 class, char *msg, uint code, uint subcode, byte *data, uint len)
 {
-  byte argbuf[256+16], *t = argbuf;
+  byte argbuf[256], *t = argbuf;
   uint i;
 
   /* Don't report Cease messages generated by myself */
@@ -3152,8 +2929,8 @@ bgp_log_error(struct bgp_proto *p, u8 class, char *msg, uint code, uint subcode,
 
   if (len)
     {
-      /* Bad peer AS / unacceptable hold time - print the value as decimal number */
-      if ((code == 2) && ((subcode == 2) || (subcode == 6)) && ((len == 2) || (len == 4)))
+      /* Bad peer AS - we would like to print the AS */
+      if ((code == 2) && (subcode == 2) && ((len == 2) || (len == 4)))
 	{
 	  t += bsprintf(t, ": %u", (len == 2) ? get_u16(data) : get_u32(data));
 	  goto done;
@@ -3244,8 +3021,6 @@ bgp_rx_packet(struct bgp_conn *conn, byte *pkt, uint len)
   byte type = pkt[18];
 
   DBG("BGP: Got packet %02x (%d bytes)\n", type, len);
-  conn->bgp->stats.rx_messages++;
-  conn->bgp->stats.rx_bytes += len;
 
   if (conn->bgp->p.mrtdump & MD_MESSAGES)
     bgp_dump_message(conn, pkt, len);
