@@ -1,8 +1,6 @@
 /*
  *	BIRD -- Router Advertisement
  *
- *	(c) 2011--2019 Ondrej Zajicek <santiago@crfreenet.org>
- *	(c) 2011--2019 CZ.NIC z.s.p.o.
  *
  *	Can be freely distributed and used under the terms of the GNU GPL.
  */
@@ -37,14 +35,18 @@
  * case of the specified trigger prefix was changed.
  *
  * Supported standards:
- * RFC 4861 - main RA standard
- * RFC 4191 - Default Router Preferences and More-Specific Routes
- * RFC 6106 - DNS extensions (RDDNS, DNSSL)
+ * - RFC 4861 - main RA standard
+ * - RFC 4191 - Default Router Preferences and More-Specific Routes
+ * - RFC 6106 - DNS extensions (RDDNS, DNSSL)
  */
 
 static void radv_prune_prefixes(struct radv_iface *ifa);
 static void radv_prune_routes(struct radv_proto *p);
 
+/* Invalidate cached RA packet */
+static inline void radv_invalidate(struct radv_iface *ifa)
+{ ifa->plen = 0; }
+
 static void
 radv_timer(timer *tm)
 {
@@ -54,13 +56,16 @@ radv_timer(timer *tm)
 
   RADV_TRACE(D_EVENTS, "Timer fired on %s", ifa->iface->name);
 
+  if (ifa->valid_time <= now)
+    radv_invalidate(ifa);
+
   if (ifa->prune_time <= now)
     radv_prune_prefixes(ifa);
 
   if (p->prune_time <= now)
     radv_prune_routes(p);
 
-  radv_send_ra(ifa, IPA_NONE);
+  radv_send_ra(ifa);
 
   /* Update timer */
   ifa->last = now;
@@ -391,12 +396,12 @@ radv_net_match_trigger(struct radv_config *cf, net *n)
 }
 
 int
-radv_preexport(struct channel *C, rte *new)
+radv_preexport(struct proto *P, rte **new, struct linpool *pool UNUSED)
 {
   // struct radv_proto *p = (struct radv_proto *) P;
-  struct radv_config *cf = (struct radv_config *) (C->proto->cf);
+  struct radv_config *cf = (struct radv_config *) (P->cf);
 
-  if (radv_net_match_trigger(cf, new->net))
+  if (radv_net_match_trigger(cf, (*new)->net))
     return RIC_PROCESS;
 
   if (cf->propagate_routes)
@@ -555,7 +560,7 @@ radv_check_active(struct radv_proto *p)
     return 1;
 
   struct channel *c = p->p.main_channel;
-  return rt_examine(c->table, &cf->trigger, c, c->out_filter);
+  return rt_examine(c->table, &cf->trigger, &p->p, c->out_filter);
 }
 
 static void
@@ -564,7 +569,7 @@ radv_postconfig(struct proto_config *CF)
   // struct radv_config *cf = (void *) CF;
 
   /* Define default channel */
-  if (! proto_cf_main_channel(CF))
+  if (EMPTY_LIST(CF->channels))
     channel_config_new(NULL, net_label[NET_IP6], NET_IP6, CF);
 }
 
@@ -622,7 +627,7 @@ radv_iface_shutdown(struct radv_iface *ifa)
   if (ifa->sk)
   {
     radv_invalidate(ifa);
-    radv_send_ra(ifa, IPA_NONE);
+    radv_send_ra(ifa);
   }
 }
 
@@ -740,7 +745,7 @@ radv_pref_str(u32 pref)
 
 /* The buffer has some minimal size */
 static int
-radv_get_attr(const eattr *a, byte *buf, int buflen UNUSED)
+radv_get_attr(eattr *a, byte *buf, int buflen UNUSED)
 {
   switch (a->id)
   {
@@ -771,9 +776,3 @@ struct protocol proto_radv = {
   .get_status =		radv_get_status,
   .get_attr =		radv_get_attr
 };
-
-void
-radv_build(void)
-{
-  proto_build(&proto_radv);
-}
